/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/types.h>
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#include <sys/mman.h>
#include <pthread.h>
#include <sched.h>
#include <math.h>
//#include <linux/sched/types.h>

struct sched_attr {
  __u32 size;

  __u32 sched_policy;
  __u64 sched_flags;

  /* SCHED_NORMAL, SCHED_BATCH */
  __s32 sched_nice;

  /* SCHED_FIFO, SCHED_RR */
  __u32 sched_priority;

  /* SCHED_DEADLINE (nsec) */
  __u64 sched_runtime;
  __u64 sched_deadline;
  __u64 sched_period;

  /* Utilization hints */
  __u32 sched_util_min;
  __u32 sched_util_max;
};

int sched_setattr(pid_t pid,
                  const struct sched_attr *attr,
                  unsigned int flags)
{
  return syscall(__NR_sched_setattr, pid, attr, flags);
}

int sched_getattr(pid_t pid,
                  struct sched_attr *attr,
                  unsigned int size,
                  unsigned int flags)
{
  return syscall(__NR_sched_getattr, pid, attr, size, flags);
}

struct task {
  pid_t id;
  unsigned long wcet;
  unsigned long period;
};

#define TASK_NUM 2
#define ARGS_NUM 2
#define DURATION_SEC 3
#define MSEC2NSEC(msec) ((msec) * 1000 * 1000L)
#define SEC2NSEC(sec) ((sec) * 1000 * 1000 * 1000L)
#define SAFETY_RATIO 0.8

void timespec_diff(struct timespec *result,
                   const struct timespec *end, const struct timespec *begin)
{
  if (end->tv_nsec - begin->tv_nsec < 0) {
    result->tv_sec = end->tv_sec - begin->tv_sec - 1;
    result->tv_nsec = end->tv_nsec - begin->tv_nsec + 1000000000;
  } else {
    result->tv_sec = end->tv_sec - begin->tv_sec;
    result->tv_nsec = end->tv_nsec - begin->tv_nsec;
  }
}

void *do_edf(void *args)
{
  struct timespec *tp = (struct timespec *)((void **) args)[0];
  struct task *task = (struct task *)((void **) args)[1];
  struct timespec now, next = {tp->tv_sec, tp->tv_nsec};
  struct sched_attr attr, attr2;
  struct timespec th_now, th_next, diff;
  int ret;
  unsigned int flags = 0;
  int job = 1;

  attr.size = sizeof(attr);
  attr.sched_policy = SCHED_DEADLINE;
  attr.sched_flags = 0;
  attr.sched_nice = 0;
  attr.sched_priority = 0; // not used

  attr.sched_runtime = MSEC2NSEC(task->wcet);
  attr.sched_period = attr.sched_deadline = MSEC2NSEC(task->period);

  attr.sched_util_min = 0;
  attr.sched_util_max = ceil((double)(task->wcet * 100) / task->period);

  if ((ret = sched_setattr(0, &attr, flags)) < 0) {
    perror("sched_setattr");
    exit(1);
  }

  if ((ret = sched_getattr(0, &attr2, sizeof(struct sched_attr), flags)) < 0) {
    perror("sched_getattr");
    exit(2);
  }

  if (attr2.sched_policy == SCHED_DEADLINE) {
    printf("sched_policy is SCHED_DEADLINE.\n");
  } else {
    fprintf(stderr, "sched_policy is not SCHED_DEADLINE.\n");
    exit(3);
  }

  printf("sleep until begin time: cpu = %d\n", sched_getcpu());

  // sleep until begin time.
  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next, NULL);

  do {
    printf("SCHED_DEADLINE thread is released [id = %d]: job = %d\n", task->id,
           job++);
    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &th_now);

    do {
      clock_gettime(CLOCK_THREAD_CPUTIME_ID, &th_next);
      timespec_diff(&diff, &th_next, &th_now);
    } while (SEC2NSEC(diff.tv_sec) + diff.tv_nsec
             <= MSEC2NSEC(task->wcet) * SAFETY_RATIO);

    sched_yield();

    if (clock_gettime(CLOCK_MONOTONIC, &now) != 0) {
      perror("clock_getitme");
      exit(4);
    }
  } while (now.tv_sec <= tp->tv_sec + DURATION_SEC);

  return NULL;
}

int main(int argc, char *argv[])
{
  pthread_t thread;
  struct task task;
  struct timespec tp;
  void *args[ARGS_NUM];

  if (argc != 4) {
    fprintf(stderr, "Usage: %s id wcet period\n", argv[0]);
    exit(1);
  }

  task.id = atoi(argv[1]);
  task.wcet = atoi(argv[2]);
  task.period = atoi(argv[3]);

  if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1) {
    perror("mlockall");
    exit(2);
  }

  if (clock_gettime(CLOCK_MONOTONIC, &tp) != 0) {
    perror("clock_getitme");
    exit(3);
  }

  // begin time is current time + 1.X [s] (at least 1 second waiting from current time).
  tp.tv_sec += 2;
  tp.tv_nsec = 0;

  args[0] = &tp;
  args[1] = &task;
  pthread_create(&thread, NULL, do_edf, args);

  pthread_join(thread, NULL);

  if (munlockall() == -1) {
    perror("mlockall");
    exit(4);
  }

  return 0;
}
