/*
 * Author: Hiroyuki Chishiro
 * License: 2-Clause BSD
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/types.h>
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#include <sys/mman.h>
#include <pthread.h>
#include <math.h>

struct sched_attr {
  __u32 size;

  __u32 sched_policy;
  __u64 sched_flags;

  /* SCHED_NORMAL, SCHED_BATCH */
  __s32 sched_nice;

  /* SCHED_FIFO, SCHED_RR */
  __u32 sched_priority;

  /* SCHED_DEADLINE (nsec) */
  __u64 sched_runtime;
  __u64 sched_deadline;
  __u64 sched_period;

  /* Utilization hints */
  __u32 sched_util_min;
  __u32 sched_util_max;
};

int sched_setattr(pid_t pid,
                  const struct sched_attr *attr,
                  unsigned int flags)
{
  return syscall(__NR_sched_setattr, pid, attr, flags);
}

int sched_getattr(pid_t pid,
                  struct sched_attr *attr,
                  unsigned int size,
                  unsigned int flags)
{
  return syscall(__NR_sched_getattr, pid, attr, size, flags);
}

struct task {
  pid_t id;
  unsigned long wcet;
  unsigned long period;
};

#define TASK_NUM 2
#define ARGS_NUM 2
#define DURATION_SEC 3
#define MSEC2NSEC(msec) ((msec) * 1000 * 1000L)
#define SEC2NSEC(sec) ((sec) * 1000 * 1000 * 1000L)
#define SAFETY_RATIO 0.8

int migrate_thread_to_cpu(pid_t tid, int target_cpu)
{
  cpu_set_t *cpu_set;
  size_t size;
  int nr_cpus;
  int ret;

  if ((nr_cpus = get_nprocs()) == -1) {
    perror("get_nprocs()");
    exit(1);
  }

  if (target_cpu < 0 || target_cpu >= nr_cpus) {
    fprintf(stderr, "Error: target_cpu %d is in [0,%d).\n", target_cpu, nr_cpus);
    exit(2);
  }

  if ((cpu_set = CPU_ALLOC(nr_cpus)) == NULL) {
    perror("CPU_ALLOC");
    exit(3);
  }

  size = CPU_ALLOC_SIZE(nr_cpus);
  CPU_ZERO_S(size, cpu_set);
  CPU_SET_S(target_cpu, size, cpu_set);

  if ((ret = sched_setaffinity(tid, size, cpu_set)) == -1) {
    perror("sched_setaffinity");
    exit(4);
  }

  CPU_FREE(cpu_set);

  return ret;
}

void set_next_release_time(struct timespec *tp, struct task *task)
{
  tp->tv_nsec += MSEC2NSEC(task->period);

  while (tp->tv_nsec >= SEC2NSEC(1)) {
    tp->tv_nsec -= SEC2NSEC(1);
    tp->tv_sec++;
  }
}

void timespec_diff(struct timespec *result,
                   const struct timespec *end, const struct timespec *begin)
{
  if (end->tv_nsec - begin->tv_nsec < 0) {
    result->tv_sec = end->tv_sec - begin->tv_sec - 1;
    result->tv_nsec = end->tv_nsec - begin->tv_nsec + 1000000000;
  } else {
    result->tv_sec = end->tv_sec - begin->tv_sec;
    result->tv_nsec = end->tv_nsec - begin->tv_nsec;
  }
}

void *do_rm(void *args)
{
  struct timespec *tp = (struct timespec *)((void **) args)[0];
  struct task *task = (struct task *)((void **) args)[1];
  struct timespec now, next = {tp->tv_sec, tp->tv_nsec};
  struct sched_attr attr, attr2;
  struct timespec th_now, th_next, diff;
  int ret;
  unsigned int flags = 0;
  int job = 1;

  attr.size = sizeof(attr);
  attr.sched_policy = SCHED_FIFO;
  attr.sched_flags = 0;
  attr.sched_nice = 0;
  attr.sched_priority = sched_get_priority_max(SCHED_FIFO) - task->id;

  attr.sched_runtime = attr.sched_period = attr.sched_deadline = 0; // not used

  attr.sched_util_min = 0;
  attr.sched_util_max = ceil((double)(task->wcet * 100) / task->period);

  if ((ret = sched_setattr(0, &attr, flags)) < 0) {
    perror("sched_setattr");
    exit(1);
  }

  if ((ret = sched_getattr(0, &attr2, sizeof(struct sched_attr), flags)) < 0) {
    perror("sched_getattr");
    exit(2);
  }

  if (attr2.sched_policy == SCHED_FIFO) {
    printf("sched_policy is SCHED_FIFO.\n");
  } else {
    fprintf(stderr, "sched_policy is not SCHED_FIFO.\n");
    exit(3);
  }

  printf("sleep until begin time: cpu = %d\n", sched_getcpu());

  // sleep until begin time.
  if (clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next, NULL) != 0) {
    perror("clock_nanosleep");
    exit(4);
  }

  do {
    printf("SCHED_FIFO thread is released [id = %d]: job = %d\n", task->id, job++);

    if (clock_gettime(CLOCK_THREAD_CPUTIME_ID, &th_now) != 0) {
      perror("clock_gettime");
      exit(5);
    }

    do {
      if (clock_gettime(CLOCK_THREAD_CPUTIME_ID, &th_next) != 0) {
        perror("clock_gettime");
        exit(6);
      }

      timespec_diff(&diff, &th_next, &th_now);
    } while (SEC2NSEC(diff.tv_sec) + diff.tv_nsec
             <= MSEC2NSEC(task->wcet) * SAFETY_RATIO);

    set_next_release_time(&next, task);

    if (clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next, NULL) != 0) {
      perror("clock_nanosleep");
      exit(7);
    }

    if (clock_gettime(CLOCK_MONOTONIC, &now) != 0) {
      perror("clock_getitme");
      exit(8);
    }

  } while (now.tv_sec <= tp->tv_sec + DURATION_SEC);

  return NULL;
}

int main(void)
{
  pthread_t threads[TASK_NUM];
  struct task tasks[TASK_NUM] = {{1, 600, 1000}, {2, 500, 1500}};
  struct timespec tp;
  void *args[TASK_NUM][ARGS_NUM];
  int i;
  pid_t tid = gettid();

  // migrate thread to CPU 0.
  migrate_thread_to_cpu(tid, 0);

  if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1) {
    perror("mlockall");
    exit(1);
  }

  if (clock_gettime(CLOCK_MONOTONIC, &tp) != 0) {
    perror("clock_getitme");
    exit(2);
  }

  // begin time is current time + 1.X [s] (at least 1 second waiting from current time).
  tp.tv_sec += 2;
  tp.tv_nsec = 0;

  for (i = 0; i < TASK_NUM; i++) {
    args[i][0] = &tp;
    args[i][1] = &tasks[i];
    pthread_create(&threads[i], NULL, do_rm, args[i]);
  }

  for (i = 0; i < TASK_NUM; i++) {
    pthread_join(threads[i], NULL);
  }

  if (munlockall() == -1) {
    perror("mlockall");
    exit(3);
  }

  return 0;
}
